# -*- coding: utf-8 -*-
# (c) 2018-2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import json
import logging

from munch import munchify
from influxdb import InfluxDBClient

from luftdatenpumpe.util import sanitize_dbsymbol

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

log = logging.getLogger(__name__)


class InfluxDBStorage:

    capabilities = ['readings']

    def __init__(self, dsn, network=None, dry_run=False):

        self.dry_run = dry_run

        self.network = network or 'default'
        self.measurement = sanitize_dbsymbol(self.network) + '_readings'

        self.buffer = []
        self.is_udp = False

        # TODO: Use timeout value from settings.
        dsn_parts = urlparse(dsn)
        if dsn_parts.scheme.startswith('udp+'):
            self.is_udp = True
            self.db = InfluxDBClient.from_dsn(dsn, udp_port=dsn_parts.port, timeout=5)
        else:
            self.db = InfluxDBClient.from_dsn(dsn, timeout=5)
            self.db.create_database(self.db._database)

    def emit(self, reading):
        """
        Obtain a reading and store into InfluxDB.

        Ingress
        =======
        A reading coming in here looks like::

            {
              "station": {
                "station_id": 11391,
                "position": {
                  "latitude": 50.846,
                  "longitude": 4.414,
                  "altitude": 78.7,
                  "country": "BE",
                  "exact_location": 0,
                  "geohash": "u151k2npmtku"
                }
              },
              "observations": [
                {
                  "meta": {
                    "timestamp": "2019-04-24T18:08:53Z",
                    "sensor_id": 22448,
                    "sensor_type": "DHT22"
                  },
                  "data": {
                    "humidity": 91.8,
                    "temperature": 13.4
                  }
                }
              ]
            }


        Egress
        ======
        The designated output format suitable for feeding into InfluxDB looks like::

            {
              "measurement": "readings",
              "tags": {
                "station_id": 28,
                "sensor_id": 658
              },
              "time": "2018-12-12T02:21:08Z",
              "fields": {
                "P1": 13.35,
                "P2": 8.05
              }
            }

        """

        # Debugging
        if log.getEffectiveLevel() is logging.DEBUG:
            log.debug('Obtained reading for InfluxDB\n%s', json.dumps(reading, indent=2))

        # Build InfluxDB records.
        for observation in reading.observations:

            record = munchify({
                "measurement": self.measurement,
                "tags": {
                    "station_id": reading.station.station_id,
                },
            })

            if 'sensor_id' in observation.meta:
                record.tags.sensor_id = observation.meta.sensor_id

            if 'position' in reading.station and 'geohash' in reading.station.position:
                record.tags.geohash = reading.station.position.geohash

            record.time = observation.meta.timestamp
            record.fields = observation.data

            # Debugging.
            if log.getEffectiveLevel() is logging.DEBUG:
                log.debug('Emitting record to InfluxDB\n%s', json.dumps(record, indent=2))

            # Store into buffer.
            self.buffer.append(record)

        if self.is_udp and len(self.buffer) > 50:
            self.flush()

    def flush(self, final=False):
        # Store into database.
        #print('Writing points:', len(self.buffer))
        self.db.write_points(self.buffer)
        self.buffer = []



# -*- coding: utf-8 -*-
# (c) 2017,2018 Andreas Motl <andreas@hiveeyes.org>
# (c) 2017,2018 Richard Pobering <richard@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import os
import re
import sys
import glob
import json
import logging
import traceback
import unicodedata
from itertools import zip_longest

from six import StringIO
from docopt import docopt
from munch import Munch, munchify
from collections import OrderedDict


log = logging.getLogger(__name__)


class Application:

    def __init__(self, name=None, version=None, docopt_recipe=None):

        self.name = name
        self.version = version
        self.docopt_recipe = docopt_recipe

        self.settings = Munch()
        self.options = Munch()

        self.initialize()

    def initialize(self):

        # Honor "LDA_" environment variables.
        argv = self.get_argv_with_environment()

        #print(argv)

        # Parse command line arguments.
        self.options = normalize_options(docopt(self.docopt_recipe, argv=argv, version=f'{self.name} {self.version}'))

        # Setup logging.
        self.setup_logging()

    def setup_logging(self):
        debug = self.options.get('debug')
        log_level = logging.INFO
        if debug:
            log_level = logging.DEBUG

        setup_logging(log_level)

    def log_options(self):
        # Debugging
        log.info('Options: {}'.format(json.dumps(self.options, indent=4)))

    def get_argv_with_environment(self):

        envvar_prefix = 'LDP_'

        argv = sys.argv[1:]

        def search_argv_option(barename):
            optname = '--' + barename
            for arg in argv:
                if optname in arg:
                    return True
            return False

        for name, value in os.environ.items():
            if name.startswith(envvar_prefix):
                optname = name.replace(envvar_prefix, '').lower()
                if not search_argv_option(optname):
                    option = f'--{optname}={value}'
                    argv.append(option)

        return argv


def setup_logging(level=logging.INFO):
    log_format = '%(asctime)-15s [%(name)-36s] %(levelname)-7s: %(message)s'
    logging.basicConfig(
        format=log_format,
        stream=sys.stderr,
        level=level)

    # TODO: Control debug logging of HTTP requests through yet another commandline option "--debug-http" or "--debug-requests"
    requests_log = logging.getLogger('requests')
    requests_log.setLevel(logging.WARN)


def normalize_options(options):
    normalized = {}
    for key, value in options.items():
        key = key.strip('--<>')
        #key = key.replace('--<>', '')
        normalized[key] = value
    return munchify(normalized)


def read_list(data, separator=u','):
    if data is None:
        return []
    result = list(map(lambda x: x.strip(), data.split(separator)))
    if len(result) == 1 and not result[0]:
        result = []
    return result


def read_pairs(data, listsep=u',', pairsep=u'='):
    if data is None:
        return {}
    pairs = list(map(lambda x: x.strip(), data.split(listsep)))
    if len(pairs) == 1 and not pairs[0]:
        return {}
    result = OrderedDict()
    for pair in pairs:
        key, value = pair.split(pairsep)
        result[key] = value
    return result


def exception_traceback(exc_info=None):
    """
    Return a string containing a traceback message for the given
    exc_info tuple (as returned by sys.exc_info()).

    from setuptools.tests.doctest
    """

    if not exc_info:
        exc_info = sys.exc_info()

    # Get a traceback message.
    excout = StringIO()
    exc_type, exc_val, exc_tb = exc_info
    traceback.print_exception(exc_type, exc_val, exc_tb, file=excout)
    return excout.getvalue()


def to_list(obj):
    """Convert an object to a list if it is not already one"""
    if not isinstance(obj, (list, tuple)):
        obj = [obj, ]
    return obj


def invalidate_dogpile_cache(cache_instance, hard=True):
    """
    The default invalidation system works by setting
    a current timestamp (using ``time.time()``)
    representing the "minimum creation time" for
    a value.  Any retrieved value whose creation
    time is prior to this timestamp
    is considered to be stale.  It does not
    affect the data in the cache in any way, and is
    **local to this instance of :class:`.CacheRegion`.**

    .. warning::

        The :meth:`.CacheRegion.invalidate` method's default mode of
        operation is to set a timestamp **local to this CacheRegion
        in this Python process only**.   It does not impact other Python
        processes or regions as the timestamp is **only stored locally in
        memory**.  To implement invalidation where the
        timestamp is stored in the cache or similar so that all Python
        processes can be affected by an invalidation timestamp, implement a
        custom :class:`.RegionInvalidationStrategy`.

    The method supports both "hard" and "soft" invalidation
    options.  With "hard" invalidation,
    :meth:`.CacheRegion.get_or_create` will force an immediate
    regeneration of the value which all getters will wait for.
    With "soft" invalidation, subsequent getters will return the
    "old" value until the new one is available.
    """
    cache_instance.invalidate(hard=hard)


def find_files_glob(pattern):
    return sorted(glob.glob(pattern, recursive=True))


def find_files_walk(path, suffix):
    for root, subdirs, files in os.walk(path):
        # https://python-forum.io/Thread-why-i-don-t-like-os-walk?pid=35718#pid35718
        subdirs[:] = sorted(subdirs)  # sorts the subdirs
        for name in sorted(files):
            realname = os.path.join(root, name)
            if realname.endswith(suffix):
                yield realname

find_files = find_files_glob


def sanitize_dbsymbol(symbol):
    return symbol.replace('-', '_')


def slugify(value):
    """
    Slugify function from django.utils.text, slightly modified.

    Converts to lowercase, removes non-word characters (alphanumerics and
    underscores) and converts spaces to hyphens. Also strips leading and
    trailing whitespace.

    Update 2019-04-26.
    Account for strings like "wind speed (scalar)" being
    converted to things like "wind-speed-scalar-".
    So, also strip these leftovers.

    https://stackoverflow.com/questions/5574042/string-slugification-in-python/27264385#27264385
    """
    value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')
    value = re.sub('[^\w\s-]', '-', value).strip().lower()
    value = re.sub('[-\s]+', '-', value)
    value = value.strip('-')
    return value


def is_nan(value):
    return value is None or str(value).lower() == 'nan'


def run_once(f):
    # https://stackoverflow.com/questions/4103773/efficient-way-of-having-a-function-only-execute-once-in-a-loop/4104188#4104188
    def wrapper(*args, **kwargs):
        if not wrapper.has_run:
            wrapper.has_run = True
            return f(*args, **kwargs)
    wrapper.has_run = False
    return wrapper


def grouper(iterable, n, fillvalue=None):
    """
    Collect data into fixed-length chunks or blocks
    https://docs.python.org/3/library/itertools.html#itertools-recipes
    """
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


def remove_all(data, element):
    while True:
        try:
            data.remove(element)
        except ValueError:
            break
    return data


def chunks(iterable, chunksize):
    for group in grouper(iterable, chunksize):
        yield remove_all(list(group), None)



# -*- coding: utf-8 -*-
# (c) 2017-2019 Andreas Motl <andreas@hiveeyes.org>
# (c) 2017-2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Matthias Mehldau <wetter@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import sys
import json
import logging
from docopt import docopt, DocoptExit
from luftdatenpumpe import __appname__, __version__
from luftdatenpumpe.grafana import get_artefact
from luftdatenpumpe.source import resolve_source_handler
from luftdatenpumpe.source.rdbms import stations_from_rdbms
from luftdatenpumpe.engine import LuftdatenEngine
from luftdatenpumpe.util import normalize_options, setup_logging, read_pairs, Application

log = logging.getLogger(__name__)

network_list = ['ldi', 'irceline']


def run():
    """
    Usage:
      luftdatenpumpe networks
      luftdatenpumpe stations --network=<network> [options] [--target=<target>]...
      luftdatenpumpe readings --network=<network> [options] [--target=<target>]... [--timespan=<timespan>]
      luftdatenpumpe database --network=<network> [--target=<target>]... [--create-views] [--grant-user=<username>] [--drop-data] [--drop-tables] [--drop-database]
      luftdatenpumpe grafana --network=<network> --kind=<kind> --name=<name> [--variables=<variables>] [--fields=<fields>]
      luftdatenpumpe --version
      luftdatenpumpe (-h | --help)

    Options:
      --network=<network>           Which sensor network/database to use.
                                    Inquire available networks by running "luftdatenpumpe networks".
      --source=<source>             Data source, either "api" or "file://" [default: api].
      --country=<countries>         Filter data by given country codes, comma-separated.
      --station=<stations>          Filter data by given location ids, comma-separated.
      --sensor=<sensors>            Filter data by given sensor ids, comma-separated.
      --sensor-type=<sensor-types>  Filter data by given sensor types, comma-separated.
      --timespan=<timespan>         Filter readings by time range, only for SOS API (e.g. IRCELINE).
      --reverse-geocode             Compute geographical address using the Nominatim reverse geocoder
      --target=<target>             Data output target
      --disable-nominatim-cache     Disable Nominatim reverse geocoder cache
      --progress                    Show progress bar
      --version                     Show version information
      --dry-run                     Skip publishing to MQTT bus
      --debug                       Enable debug messages
      -h --help                     Show this screen


    Network list example:

      # Display list of supported sensor networks
      luftdatenpumpe networks

    Basic station list examples:

      # Display metadata for given countries in JSON format
      luftdatenpumpe stations --network=ldi --database=ldi --country=BE,NL,LU

      # Display metadata for given stations in JSON format, with reverse geocoding
      luftdatenpumpe stations --network=ldi --station=28,1071 --reverse-geocode


    Heads up!

      From now on, let's pretend we always want to operate on data coming from the
      sensor network "luftdaten.info". which is identified by "--network=ldi". To
      make this more convenient, we use an environment variable to signal this
      to subsequent invocations of "luftdatenpumpe" by running::

        export LDP_NETWORK=ldi

    Further examples:

      # Display metadata for given sensors in JSON format, with reverse geocoding
      luftdatenpumpe stations --sensor=28,1071 --reverse-geocode

      # Display list of stations in JSON format made of value/text items, suitable for use as a Grafana JSON data source
      luftdatenpumpe stations --station=28,1071 --reverse-geocode --target=json.grafana.vt+stream://sys.stdout

      # Display list of stations in JSON format made of key/name items, suitable for use as a mapping in Grafana Worldmap Panel
      luftdatenpumpe stations --station=28,1071 --reverse-geocode --target=json.grafana.kn+stream://sys.stdout

      # Store list of stations and metadata into RDBMS database (PostgreSQL), also display on STDERR
      luftdatenpumpe stations --station=28,1071 --reverse-geocode --target=postgresql://luftdatenpumpe@localhost/weatherbase --target=json+stream://sys.stderr

      # Read station information from RDBMS database (PostgreSQL) and format for Grafana Worldmap Panel
      luftdatenpumpe stations --source=postgresql://luftdatenpumpe@localhost/weatherbase --target=json.grafana.kn+stream://sys.stdout


    Live data examples (InfluxDB):

      # Store into InfluxDB running on "localhost"
      luftdatenpumpe readings --station=28,1071 --target=influxdb://luftdatenpumpe@localhost/luftdaten_info

      # Store into InfluxDB, with UDP
      luftdatenpumpe readings --station=28,1071 --target=udp+influxdb://localhost:4445/luftdaten_info

      # Store into InfluxDB, with authentication
      luftdatenpumpe readings --station=28,1071 --target=influxdb://luftdatenpumpe@localhost/luftdaten_info


    LDI CSV archive data examples (InfluxDB):

      # Mirror archive of luftdaten.info, limiting to 2015 only
      wget --mirror --continue --no-host-directories --directory-prefix=/var/spool/archive.luftdaten.info --accept-regex='2015' http://archive.luftdaten.info/

      # Ingest station information from CSV archive files, store into PostgreSQL
      luftdatenpumpe stations --network=ldi --source=file:///var/spool/archive.luftdaten.info --target=postgresql://luftdatenpumpe@localhost/weatherbase --reverse-geocode --progress

      # Ingest readings from CSV archive files, store into InfluxDB
      luftdatenpumpe readings --network=ldi --source=file:///var/spool/archive.luftdaten.info --station=483 --sensor=988 --target=influxdb://luftdatenpumpe@localhost/luftdaten_info --progress

      # Ingest most early readings
      luftdatenpumpe readings --network=ldi --source=file:///var/spool/archive.luftdaten.info/2015-10-*

      # Ingest most early PMS sensors
      luftdatenpumpe readings --network=ldi --source=file:///var/spool/archive.luftdaten.info/2017-1*/*pms*.csv


    Live data examples (MQTT):

      # Publish data to topic "luftdaten.info" at MQTT broker running on "localhost"
      luftdatenpumpe readings --station=28,1071 --target=mqtt://localhost/luftdaten.info

      # MQTT publishing, with authentication
      luftdatenpumpe readings --station=28,1071 --target=mqtt://username:password@localhost/luftdaten.info


    Combined examples:

      # Write stations to STDERR and PostgreSQL
      luftdatenpumpe stations --station=28,1071 --target=json+stream://sys.stderr --target=postgresql://luftdatenpumpe@localhost/weatherbase

      # Write readings to STDERR, MQTT and InfluxDB
      luftdatenpumpe readings --station=28,1071 --target=json+stream://sys.stderr --target=mqtt://localhost/luftdaten.info --target=influxdb://luftdatenpumpe@localhost/luftdaten_info


    """

    # Bootstrap application.
    application = Application(name=__appname__, version=__version__, docopt_recipe=run.__doc__)

    options = application.options

    # 1. Run some sanity checks on ingress parameters.
    if options.networks:
        log.info('List of available networks: %s', network_list)
        sys.exit(0)
    check_options(options)

    # 2. Dispatch to maintenance targets.
    run_maintenance(options)

    # 3. Dispatch to data processing targets.
    run_engine(options)


def run_maintenance(options):

    # A. Maintenance targets
    # Create database view "ldi_network" spanning all "ldi_*" tables.
    if options['database']:

        if not options['target']:
            message = 'No target for database operation given'
            log.error(message)
            raise NotImplementedError(message)

        engine = get_engine(options)
        for target in options['target']:
            if target.startswith('postgresql:'):
                handler = engine.resolve_target_handler(target)

                try:

                    if options['drop-data']:
                        handler.drop_data()

                    if options['drop-tables']:
                        handler.drop_tables()

                    if options['drop-database']:
                        handler.drop_database()

                    if options['create-views']:
                        log.info('Creating database views')
                        handler.create_views()

                    if options['grant-user']:
                        log.info('Granting read privileges to "%s"', options['grant-user'])
                        handler.grant_read_privileges(options['grant-user'])

                except Exception as ex:
                    log.exception('Database operation failed. Reason: %s', ex, exc_info=False)

            else:
                log.warning('Can not run database operation on "%s"', target)

        sys.exit()

    # Generate JSON for Grafana datasource or dashboard and exit.
    elif options['grafana']:
        options.variables = read_pairs(options.variables)
        options.fields = read_pairs(options.fields)
        thing = get_artefact(options.kind, options.name, variables=options.variables, fields=options.fields)
        print(thing)
        sys.exit()


def check_options(options):

    # A. Sanity checks
    if not options.network:
        message = '--network parameter missing'
        log.error(message)
        raise DocoptExit(message)

    # A.2 Resolve data source handler class from network identifier.
    options.network = options.network.lower()

    # A.3 Resolve data domain (stations vs. readings).
    options.domain = None
    for flavor in ['stations', 'readings']:
        if options[flavor]:
            options.domain = flavor
            break


def get_engine(options):

    # Create and run output processing engine.
    log.info(f'Will publish data to {options.target}')
    engine = LuftdatenEngine(
        network=options.network,
        domain=options.domain,
        targets=options.target,
        progressbar=options.progress,
        dry_run=options['dry-run'],
    )

    return engine


def get_data(options):

    pump = resolve_source_handler(options)

    # Acquire data.
    if options.domain == 'stations':
        log.info(f'Acquiring list of stations from network "{options.network}" with source "{options.source}"')

        if options.source.startswith('postgresql://'):
            data = stations_from_rdbms(options.source, options.network)

        else:
            data = pump.get_stations()

        log.info(f'Acquired #{len(data)} stations')

    elif options.domain == 'readings':
        log.info(f'Acquiring readings from network "{options.network}" with source "{options.source}"')
        data = pump.get_readings()

    else:
        raise DocoptExit('Subcommand not implemented')

    # Sanity checks.
    if data is None:
        log.error('No data to process')
        sys.exit(2)

    return data


def run_engine(options):
    data = get_data(options)
    engine = get_engine(options)
    engine.process(data)



# -*- coding: utf-8 -*-
# (c) 2018-2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import logging

from munch import Munch

from luftdatenpumpe.geo import disable_nominatim_cache
from luftdatenpumpe.source.luftdaten_info import LuftdatenPumpe
from luftdatenpumpe.source.irceline import IrcelinePumpe
from luftdatenpumpe.util import read_list

log = logging.getLogger(__name__)


def resolve_source_handler(options, dry_run=False):
    handler = None

    # TODO: Add more data sources.
    if options.network == 'ldi':
        datapump_class = LuftdatenPumpe

    elif options.network == 'irceline':
        datapump_class = IrcelinePumpe

    else:
        message = f'Network "{options.network}" not implemented'
        log.error(message)
        raise NotImplementedError(message)


    # B. Data processing targets

    # Optionally apply filters by country code, station id and/or sensor id
    filter = Munch()

    # Lists of Integers.
    for filter_name in ['station', 'sensor']:
        if options[filter_name]:
            filter[filter_name] = list(map(int, read_list(options[filter_name])))

    # Lists of lower-case Strings.
    for filter_name in ['sensor-type']:
        if options[filter_name]:
            filter[filter_name] = list(map(str.lower, read_list(options[filter_name])))

    # Lists of upper-case Strings.
    for filter_name in ['country']:
        if options[filter_name]:
            filter[filter_name] = list(map(str.upper, read_list(options[filter_name])))

    if options.timespan:
        filter.timespan = options.timespan

    if filter:
        log.info('Applying filter: {}'.format(filter))

    # Default output target is STDOUT.
    if not options['target']:
        options['target'] = ['json+stream://sys.stdout']

    # Optionally disable Nominatim cache.
    if options['disable-nominatim-cache']:
        # Invalidate the Nominatim cache; this applies only for this session, it will _not_ _purge_ all data at once.
        disable_nominatim_cache()

    # The main workhorse.
    pump = datapump_class(
        source=options['source'],
        filter=filter,
        reverse_geocode=options['reverse-geocode'],
        progressbar=options['progress'],
        dry_run=options['dry-run'],
    )

    return pump



# -*- coding: utf-8 -*-
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import json
import logging
from jinja2 import Template
from pkg_resources import resource_string


log = logging.getLogger(__name__)


def get_artefact(kind, name, variables=None, fields=None):
    log.info(f'Generating Grafana artefact '
             f'kind={kind}, name={name}, variables={json.dumps(variables)}, fields={json.dumps(fields)}')
    variables = variables or {}
    fields = fields or {}
    variables['fields'] = fields
    filename = '{}-{}.json'.format(kind, name)
    template = Template(load_file(filename))
    payload = template.render(variables)
    return payload


def load_file(filename):
    return resource_string('luftdatenpumpe.grafana', filename).decode('utf-8')



# -*- coding: utf-8 -*-
# (c) 2018-2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import json
import types
import logging


log = logging.getLogger(__name__)


def json_formatter(data):
    if isinstance(data, types.GeneratorType):
        data = list(data)
    return json.dumps(data, indent=2)


def json_grafana_formatter_base(stations):
    entries = []
    for station in stations:
        if 'name' in station:
            station_name = station.name
        else:
            station_name = u'Station #{}, {}'.format(station.station_id, station.position.country)
        entry = {'value': station.station_id, 'text': station_name}
        entries.append(entry)
    return entries


def json_grafana_formatter_vt(stations):
    """
    Format list of stations in JSON format made of value/text items,
    suitable for use as a Grafana JSON data source.

    Example::

        luftdatenpumpe stations --source=postgresql://luftdatenpumpe@localhost/weatherbase --target=json.grafana.vt+stream://sys.stdout

    Emits items like::

        {
            "value": 22,
            "text": "Steglen, Haslach, Herrenberg, Vereinbarte  der Stadt Herrenberg, Baden-W\u00fcrttemberg, DE"
        }

    """
    entries = json_grafana_formatter_base(stations)
    return json_formatter(entries)


def json_grafana_formatter_kn(stations):
    """
    Format list of stations in JSON format made of key/name items,
    suitable for use as a mapping in Grafana Worldmap Panel.

    Example::

        luftdatenpumpe stations --source=postgresql://luftdatenpumpe@localhost/weatherbase --target=json.grafana.kn+stream://sys.stdout

    Emits items like::

        {
            "key": "22",
            "name": "Steglen, Haslach, Herrenberg, Vereinbarte  der Stadt Herrenberg, Baden-W\u00fcrttemberg, DE"
        }

    """
    entries = []
    for station in json_grafana_formatter_base(stations):
        entry = {'key': str(station['value']), 'name': station['text']}
        entries.append(entry)
    return json_formatter(entries)



# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()

requires = [

    # Core
    'six==1.12.0',
    'appdirs==1.4.3',
    'docopt==0.6.2',
    'munch==2.3.2',
    'tqdm==4.31.1',

    # Acquisition
    'requests==2.21.0',
    'tablib==0.13.0',

    # Caching
    'requests-cache==0.5.0',
    'dogpile.cache==0.7.1',
    'redis==3.2.1',

    # Adapters
    'paho-mqtt==1.4.0',
    'dataset==1.1.2',
    'psycopg2-binary==2.8.2',
    'GeoAlchemy2==0.6.1',
    'influxdb==5.2.2',
    'SQLAlchemy-Utils==0.33.11',

    # Date/Time
    'rfc3339==6.0',

    # Geospatial
    'geohash2==1.1',
    'geopy==1.19.0',

    # Grafana
    'Jinja2==2.10.1',
]

extras = {
    'test': [
        'pytest==4.0.1',
        'pytest-cov==2.6.0',
    ],
}

setup(name='luftdatenpumpe',
      version='0.11.0',
      description='Process live and historical data from luftdaten.info. Filter by station-id, '
                  'sensor-id and sensor-type, apply reverse geocoding, store into TSDB and '
                  'RDBMS databases, publish to MQTT or just output as JSON.',
      long_description=README,
      license="AGPL 3, EUPL 1.2",
      classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        "Topic :: Communications",
        "Topic :: Database",
        "Topic :: Internet",
        "Topic :: Internet :: WWW/HTTP :: Indexing/Search",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: Scientific/Engineering :: Human Machine Interfaces",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Embedded Systems",
        "Topic :: Software Development :: Libraries",
        "Topic :: System :: Archiving",
        "Topic :: System :: Networking :: Monitoring",
        "Topic :: Text Processing",
        "Topic :: Utilities",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS"
        ],
      author='Andreas Motl',
      author_email='andreas@hiveeyes.org',
      url='https://github.com/hiveeyes/luftdatenpumpe',
      keywords='luftdaten.info irceline '
               'air quality particulate matter pollution '
               'feinstaub luftdaten '
               'ogc sos '
               'sensor network observation '
               'opendata data acquisition transformation export '
               'geospatial temporal timeseries '
               'http rest json api '
               'rdbms sql mysql '
               'mosquitto mqtt '
               'openstreetmap nominatim '
               'postgis postgresql influxdb grafana 52north',
      packages=find_packages(),
      include_package_data=True,
      package_data={
      },
      zip_safe=False,
      test_suite='luftdatenpumpe.test',
      install_requires=requires,
      extras_require = extras,
      tests_require=extras['test'],
      dependency_links=[
      ],
      entry_points={
          'console_scripts': [
              'luftdatenpumpe = luftdatenpumpe.commands:run',
          ],
      },
)



# -*- coding: utf-8 -*-
# (c) 2017,2018 Andreas Motl <andreas@hiveeyes.org>
# (c) 2017,2018 Richard Pobering <richard@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import re
import time
import logging
import geohash2
from copy import deepcopy
from munch import Munch
from dogpile.cache import make_region
from dogpile.cache.util import kwarg_function_key_generator, to_list
from geopy.geocoders import Nominatim

from luftdatenpumpe.util import invalidate_dogpile_cache
from . import __appname__ as APP_NAME
from . import __version__ as APP_VERSION


log = logging.getLogger(__name__)


# Configure cache for responses from Nominatim
nominatim_cache = make_region(
    function_key_generator=kwarg_function_key_generator) \
    .configure('dogpile.cache.redis')

# Configure Nominatim client
nominatim_user_agent = APP_NAME + '/' + APP_VERSION


# Maybe also add building, public_building
osm_address_fields = \
    [
        'continent',
        'country_code', 'country', 'state', 'state_district', 'county',
        'postcode',
        'post_box',
        'city', 'town', 'village', 'city_district', 'suburb', 'residential',
        'road', 'pedestrian', 'cycleway', 'footway', 'path',
        'house_number',

        'administrative', 'region', 'neighbourhood', 'industrial', 'common',
        'address26', 'address29',
    ]


def resolve_location(latitude=None, longitude=None, geohash=None, country_code=None, improve=True):

    # If only geohash is given, convert back to lat/lon.
    if latitude is None and longitude is None and geohash is not None:
        latitude, longitude = geohash_decode(geohash)

    # Run reverse geocoder.
    location = reverse_geocode(latitude, longitude, country_code)

    # Backward compatibility re. already cached objects from ``reverse_geocode``.
    if hasattr(location, 'raw'):
        location = location.raw

    if location is None or 'error' in location:
        raise ValueError('Reverse geocoding failed for lat={}, lon={}'.format(latitude, longitude))

    location = rebundle_location(location)

    return location


def rebundle_location(location):

    address = Munch()
    for field in osm_address_fields:
        if field in location['address']:
            address[field] = location['address'][field]
            del location['address'][field]

    address_more = Munch()
    for key, value in location['address'].items():
        address_more[key] = value

    del location['address']

    result = Munch(location)
    result.address = address

    if address_more:
        result.address_more = address_more

    return result


# Cache responses from Nominatim for 3 months
@nominatim_cache.cache_on_arguments(expiration_time=60 * 60 * 24 * 30 * 3)
def reverse_geocode(latitude, longitude, country_code):
    """
    Cache responses of the Nominatim reverse geocoding service.

    Uses a local version of Nomatim to get rid of
    the "one request per second" fair-use policy.

    https://wiki.openstreetmap.org/wiki/Nominatim/Installation

    This is a Python implementation for subsequently requesting::

        time http https://nominatim.hiveeyes.org/reverse lat==52.246 lon==20.898 format==json
        time http https://nominatim.openstreetmaps.org/reverse lat==52.246 lon==20.898 format==json

    """

    location = None
    if country_code == 'DE':
        location = reverse_geocode_main(latitude, longitude)

    if location is None or 'error' in location:
        location = reverse_geocode_fallback(latitude, longitude)

    if location is None or 'error' in location:
        # Must raise an exception here to let the result not be cached.
        raise ValueError('Reverse geocoding failed for lat={}, lon={}'.format(latitude, longitude))

    return location


def reverse_geocode_main(latitude, longitude):

    log.debug('Reverse geocoding I: {}'.format(locals()))

    location = None
    try:
        geolocator = Nominatim(domain='nominatim.hiveeyes.org', user_agent=nominatim_user_agent, timeout=1.25)
        position = (latitude, longitude)
        location = geolocator.reverse(position).raw

        # Sanity checks
        if location['address'] == {'country': 'Deutschland', 'country_code': 'de', 'continent': 'Europe'}:
            return None

    except Exception as ex:
        name = ex.__class__.__name__
        log.warning('Reverse geocoding I failed: {}: {}. lat={}, lon={}'.format(name, ex, latitude, longitude))

    return location


def reverse_geocode_fallback(latitude, longitude):

    log.debug('Reverse geocoding II: {}'.format(locals()))

    location = None
    try:
        # 2018-03-24
        # Nominatim expects the User-Agent as HTTP header otherwise it returns a HTTP-403.
        # This has been fixed in geopy-1.12.0.
        # https://operations.osmfoundation.org/policies/nominatim/
        # https://github.com/geopy/geopy/issues/185
        geolocator = Nominatim(user_agent=nominatim_user_agent, timeout=3.0)

        # FIXME: When using "HTTP_PROXY" from environment, use scheme='http'
        # export HTTP_PROXY=http://weather.hiveeyes.org:8912/
        # See also https://github.com/geopy/geopy/issues/263
        #geolocator = Nominatim(user_agent=nominatim_user_agent, scheme='http')

        position = (latitude, longitude)
        location = geolocator.reverse(position).raw

    except Exception as ex:
        name = ex.__class__.__name__
        log.error('Reverse geocoding II failed: {}: {}. lat={}, lon={}'.format(name, ex, latitude, longitude))

    finally:
        # Obey to fair use policy (an absolute maximum of 1 request per second).
        # https://operations.osmfoundation.org/policies/nominatim/
        time.sleep(1)

    return location


def improve_location(location):
    """
    Heuristically compensate for anomalies from upstream OSM.

    TODO:
    - Get urban vs. rural sorted out

        1. country
        2. state
        3. "Überregional": "q-region"
            - county, state_district, state
        4. "Regional": "q-hood"
            - neighbourhood vs. quarter vs. residential vs. suburb vs. city_district vs. city vs. allotments

    - How to handle "building", "public_building", "residential", "pedestrian", "kindergarten", "clothes"?
    - Also check OSM place: https://wiki.openstreetmap.org/wiki/Key:place !!!
    """

    if location is None:
        return

    address = location.address

    # Uppercase `country_code` attribute.
    if 'country_code' in address:
        address.country_code = address.country_code.upper()


    # Patch `country` field.
    if 'country_code' in address:

        # R.D. => Dominican Republic
        # Example: Guayacanes, Costámbar, Puerto Plata, DO (#109)
        if address.country_code == 'DO':
            address.country = 'Dominican Republic'

        # RP => Republic of Poland
        # Example: Polarna, Krzyki, Osiedle Krzyki-Partynice, Wrocław, dolnośląskie, PL (#1463)
        if address.country_code == 'PL':
            address.country = 'Poland'


    # Improve `city` attribute.

    # Stadtstaat handling. For all known Stadtstaaten,
    # - if `city` is missing, use `state` attribute.
    # - if `state` is missing, use `city` attribute.
    # https://de.wikipedia.org/wiki/Stadtstaat#Souver%C3%A4ne_Stadtstaaten
    stadtstaaten = [

        # Germany
        'Berlin', 'Bremen', 'Hamburg',

        # Austria
        'Wien',
    ]
    city_or_state = address.get('city', address.get('state'))
    if city_or_state in stadtstaaten:
        address.state = address.city = city_or_state

    # If `city` is missing, try a number of alternative attributes.
    city_choices = [

        # Aliases
        'village', 'town', 'county', 'suburb', 'city_district',

        # Stadtstaat
        'state',
    ]
    if 'city' not in address:
        for fieldname in city_choices:
            if fieldname in address:
                address.city = address[fieldname]
                break

    # Patch `city` field.
    if 'city' in address and address.city == 'Rgbg':
        address.city = 'Regensburg'

    if 'city' not in address:
        address.city = 'unknown'


    # Improve Stadtteil.
    # Use `residential` or `neighbourhood` for missing `suburb` attribute.
    if 'suburb' not in address and 'residential' in address:
        address.suburb = address.residential
    if 'suburb' not in address and 'neighbourhood' in address:
        address.suburb = address.neighbourhood

    # Be agnostic against road vs. path
    if 'road' not in address:
        road_choices = ['path', 'pedestrian', 'cycleway', 'footway', 'neighbourhood', 'house_number']
        for fieldname in road_choices:
            if fieldname in address:
                address.road = address[fieldname]
                break

    if 'road' not in address:
        address.road = 'unknown'


def format_address(location):
    """
    The canonical station ``name`` attribute yields a more compact display
    name than that already provided by the upstream OSM ``display_name`` attribute.

    References
    ----------
    - https://github.com/OpenCageData/address-formatting
    - https://opencagedata.com/
    - https://openaddresses.io/
    - http://results.openaddresses.io/
    - https://github.com/DenisCarriere/geocoder-geojson

    """

    location = deepcopy(location)
    address = location.address

    # Clean up `county`, 'city_district` and `suburb` fields.
    blacklist = [
        # county
        'Landkreis', 'Kreis', 'Verwaltungsgemeinschaft',

        # city_district
        'Stadtbezirk',

        # suburb: prefixes
        'Bezirksteil', 'Bezirk', 'KG',

        # suburb: values
        'Fhain',
    ]
    blacklist_fields = ['county', 'city_district', 'suburb']
    for fieldname in blacklist_fields:
        if fieldname in address:
            for black in blacklist:
                if black in address[fieldname]:
                    address[fieldname] = address[fieldname].replace(black, '').strip()

    # Build display location from components.
    address_components = ['road', 'suburb', 'city_district', 'city', 'county', 'state', 'country_code']
    address_parts = []
    starts_with_number = re.compile('^\d+')
    for field_choices in address_components:
        field_choices = to_list(field_choices)
        for fieldname in field_choices:
            if fieldname in address:
                component = address[fieldname]

                if not component or component == 'unknown':
                    continue

                if address_parts and component == address_parts[-1]:
                    continue

                if starts_with_number.match(component):
                    continue

                address_parts.append(component)
                break

    location_label = ', '.join(address_parts)
    return location_label


def geohash_encode(latitude, longitude):
    """
    Compute Geohash from coordinates.

    -- https://en.wikipedia.org/wiki/Geohash
    """
    geohash = geohash2.encode(float(latitude), float(longitude))

    # Eight characters should be fine?
    #geohash = geohash[:8]

    return geohash


def geohash_decode(geohash):
    """
    Compute coordinates from Geohash.

    -- https://en.wikipedia.org/wiki/Geohash
    """
    return geohash2.decode(geohash)


def disable_nominatim_cache():
    log.info('Disabling Nominatim cache')
    # Invalidate the Nominatim cache; this applies only for this session, it will _not_ _purge_ all data at once.
    invalidate_dogpile_cache(nominatim_cache)
